import moment from 'moment-timezone';
import { formatCurrencyNumber, getPercentChange } from '../lib/utils';
import { Bar, CsvRows, CsvHeaderRow } from '../types';

moment.tz.setDefault('America/New_York');

const MAX_PERCENT_CHANGE = -5;
const NUMBER_OF_BARS_TO_COMPARE = 60; // 1 hour

const targetLossPercent = 1;
const targetProfitPercent = 1;

export const reportHeader: CsvHeaderRow = [
  'symbol',
  'date',
  'initial drop percent',
  'profit percent',
];

export const summaryHeader: CsvHeaderRow = ['avg profit percent'];

const barAnalyzerDrasticDrop = async ({
  bars,
  symbol,
}: {
  bars: Bar[];
  symbol: string;
}): Promise<CsvRows> => {
  const rows: CsvRows = [];
  let drasticDropDate: string | undefined = undefined;
  let drasticDropPercent: number | undefined = undefined;
  let drasticDropHasClosed: boolean | undefined = undefined;
  let drasticDropPrice: number | undefined = undefined;
  let drasticDropTargetProfitPrice: number | undefined = undefined;
  let drasticDropTargetLossPrice: number | undefined = undefined;

  const reset = () => {
    drasticDropDate = undefined;
    drasticDropPercent = undefined;
    drasticDropHasClosed = undefined;
    drasticDropPrice = undefined;
    drasticDropTargetProfitPrice = undefined;
    drasticDropTargetLossPrice = undefined;
  };

  for (const [index, bar] of bars.entries()) {
    const date = moment(bar.t).format('YYYY-MM-DD');

    if (drasticDropHasClosed) {
      if (date === drasticDropDate) {
        continue;
      }

      // if we reached the end of the day and have an open position
      if (drasticDropDate && drasticDropPercent && drasticDropPrice) {
        rows.push([
          symbol,
          drasticDropDate,
          drasticDropPercent,
          getPercentChange(bar.c, drasticDropPrice),
        ]);
      }

      reset();
    }

    const barMoment = moment(bar.t);
    const hours = barMoment.hours();
    const minutes = barMoment.minutes();

    // if before 9:30
    if (hours < 9 || (hours === 9 && minutes < 30)) {
      continue;
    }

    // after 4pm
    if (hours >= 16) {
      continue;
    }

    if (index < NUMBER_OF_BARS_TO_COMPARE - 1) {
      continue;
    }

    // last x minutes of bars
    const barsToCompare = bars.slice(
      index - NUMBER_OF_BARS_TO_COMPARE + 1,
      index + 1,
    );

    const firstBarOfSlice = barsToCompare[0];
    const percentChange = getPercentChange(bar.c, firstBarOfSlice.o);

    if (!drasticDropPercent && percentChange > MAX_PERCENT_CHANGE) {
      continue;
    }

    if (!drasticDropPercent) {
      drasticDropDate = date;
      drasticDropPercent = percentChange;
      drasticDropPrice = bar.c;
      drasticDropTargetProfitPrice = bar.c * (targetProfitPercent / 100);
      drasticDropTargetLossPrice = bar.c * (targetLossPercent / 100);
    } else if (
      drasticDropDate &&
      drasticDropTargetProfitPrice &&
      drasticDropPrice &&
      drasticDropPrice >= drasticDropTargetProfitPrice
    ) {
      drasticDropHasClosed = true;
      rows.push([
        symbol,
        drasticDropDate,
        drasticDropPercent,
        getPercentChange(bar.c, drasticDropPrice),
      ]);
    } else if (
      drasticDropDate &&
      drasticDropTargetLossPrice &&
      drasticDropPrice &&
      drasticDropPrice <= drasticDropTargetLossPrice
    ) {
      drasticDropHasClosed = true;
      rows.push([
        symbol,
        drasticDropDate,
        drasticDropPercent,
        getPercentChange(bar.c, drasticDropPrice),
      ]);
    }
  }

  return rows;
};

export const barAnalyzerSummaryDrasticDrop = (rowOfRows: CsvRows): CsvRows => {
  let totalProfitPercent = 0;

  for (const rows of rowOfRows) {
    const [, , , profitPercent] = rows;
    totalProfitPercent += profitPercent as number;
  }
  return [[formatCurrencyNumber(totalProfitPercent / rowOfRows.length)]];
};

export default barAnalyzerDrasticDrop;
